<?php

namespace Drupal\xntttsv\Plugin\ExternalEntities\StorageClient;

use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Config\ConfigFactory;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\external_entities\ResponseDecoder\ResponseDecoderFactoryInterface;
use Drupal\xnttfiles\Plugin\ExternalEntities\StorageClient\FileClientBase;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * External entities storage client for TSV files.
 *
 * @ExternalEntityStorageClient(
 *   id = "xntttsv",
 *   label = @Translation("TSV Files"),
 *   description = @Translation("Retrieves and stores records in TSV files.")
 * )
 */
class TsvFiles extends FileClientBase {

  /**
   * {@inheritdoc}
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    TranslationInterface $string_translation,
    ResponseDecoderFactoryInterface $response_decoder_factory,
    ConfigFactory $config_factory,
    MessengerInterface $messenger,
    LoggerChannelInterface $logger,
    CacheBackendInterface $cache
  ) {
    parent::__construct(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $string_translation,
      $response_decoder_factory,
      $config_factory,
      $messenger,
      $logger,
      $cache
    );
    // Defaults.
    $this->fileType = 'TSV';
    $this->fileTypePlural = 'TSV';
    $this->fileTypeCap = 'TSV';
    $this->fileTypeCapPlural = 'TSV';
    $this->fileExtension = '.tsv';
  }

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    array $configuration,
    $plugin_id,
    $plugin_definition
  ) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('string_translation'),
      $container->get('external_entities.response_decoder_factory'),
      $container->get('config.factory'),
      $container->get('messenger'),
      $container->get('logger.channel.xntttsv'),
      $container->get('cache.default')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(
    array $form,
    FormStateInterface $form_state
  ) {
    $form = parent::buildConfigurationForm($form, $form_state);
    $form['headers'] = [
      '#type' => 'radios',
      '#title' => $this->t('Column headers'),
      '#options' => [
        '' => $this->t('No column headers'),
        'first' => $this->t('First (uncommented) row'),
        'comment' => $this->t('First commented row before data rows (specify a comment prefix below)'),
      ],
      '#default_value' => $this->configuration['headers'] ?? '',
    ];
    $form['comment_prefix'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Ignore lines starting with (comment prefix)'),
      '#description' => $this->t('Leave empty if there are no comment lines. Ex.: "#". Note: empty lines are always ignored.'),
      '#default_value' => $this->configuration['comment_prefix'] ?? '',
    ];
    $form['separator'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Column separator'),
      '#description' => $this->t('Leave empty to use tabulation by default. Ex.: ";" for CSV files, "\t" for GFF3 or VCF files.'),
      '#default_value' => $this->configuration['separator'] ?? '',
    ];
    $form['quoted'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Support quoted values'),
      '#description' => $this->t('If checked, separators, new lines and 2 following double quotes found between double quotes are ignored.'),
      '#default_value' => $this->configuration['quoted'] ?? FALSE,
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  protected function parseFile(string $file_path) :array {
    $data = [];
    $raw_tsv = file($file_path);
    if (!empty($raw_tsv)) {
      $trash_regex = '/^\\s*$/';
      // Prepare empty lines and comment regex.
      if (!empty($this->configuration['comment_prefix'])) {
        $comment_prefix_len = strlen($this->configuration['comment_prefix']);
        $trash_regex =
          '/^\Q'
          . $this->configuration['comment_prefix']
          . '\E|^\s*$/'
        ;
      }
      // $split_columns is a function used to split row columns.
      $sep = stripcslashes(
        empty($this->configuration['separator'])
        ? "\t"
        : $this->configuration['separator']
      );
      $split_columns = function (string $s) use ($sep) {
        // Remove EOL chars.
        $s = rtrim($s, "\n\r");
        // Check quote support.
        if (empty($this->configuration['quoted'])) {
          return explode($sep, $s);
        }
        else {
          $unquoted_values = [];
          $quote_exploded = explode($sep, $s);
          $piece = array_shift($quote_exploded);
          do {
            if (!empty($piece) && ('"' === $piece[0])) {
              // Quoted value, check quote count.
              $quote_count = substr_count($piece, '"');
              while ($quote_count % 2) {
                $next_piece = array_shift($quote_exploded);
                if (!isset($next_piece)) {
                  break;
                }
                $quote_count += substr_count($next_piece, '"');
                // Concatenate split values.
                $piece .= $sep . $next_piece;
              }
              if (substr($piece, -1) === '"') {
                // Strip quotes.
                $piece = substr($piece, 1, -1);
              }
              else {
                // Warn for format error.
                $this->logger->warning(
                  'File format error: invalid quotes in file "'
                  . $file_path
                  . '" (XNTT type: "'
                  . $this->externalEntityType->getDerivedEntityTypeId()
                  . "\"):\n"
                  . $piece
                );
              }
            }
            $unquoted_values[] = $piece;
            $piece = array_shift($quote_exploded);
          } while (isset($piece));
          return $unquoted_values;
        }
      };

      // Process header.
      $column_line = $previous_line = '';
      do {
        if (!empty($column_line)) {
          $previous_line = $column_line;
        }
        $column_line = array_shift($raw_tsv);
        // Skip comments and empty lines.
      } while (isset($column_line) && preg_match($trash_regex, $column_line));

      // Get field names.
      switch ($this->configuration['headers'] ?? '') {
        case 'first':
          $data[''] = $columns = $split_columns($column_line);
          break;

        case 'comment':
          array_unshift($raw_tsv, $column_line);
          $column_line = $previous_line ?? '';
          // Remove comment prefix (if one).
          if ($comment_prefix_len) {
            $row_prefix = substr($column_line, 0, $comment_prefix_len);
            if ($row_prefix == $this->configuration['comment_prefix']) {
              $column_line = substr($column_line, $comment_prefix_len);
            }
          }
          $data[''] = $columns = $split_columns($column_line);
          break;

        default:
          // Column index will be added after.
          $data[''] = [];
          $columns = 0;
          break;
      }
      // No data, stop here.
      if (!isset($column_line)) {
        return [];
      }
      // Iterate on entity lines.
      $idf = $this->getIdField();
      $raw_line = array_shift($raw_tsv);
      // No data, stop here.
      if (!isset($raw_line)) {
        return [];
      }
      do {
        if (preg_match($trash_regex, $raw_line)) {
          // Skip comments and empty lines.
          continue;
        }
        // Check quote support.
        if (!empty($this->configuration['quoted'])) {
          // Check if line contains quoted line break.
          $quote_count = substr_count($raw_line, '"');
          while ($quote_count % 2) {
            $next_raw_line = array_shift($raw_tsv);
            if (!isset($next_raw_line)) {
              break;
            }
            // Concatenate broken lines.
            $quote_count += substr_count($next_raw_line, '"');
            $raw_line .= $next_raw_line;
          }
        }

        // Extract value and also remove EOL chars.
        $values = $split_columns($raw_line);
        // Check how to handle columns.
        if (is_int($columns)) {
          // No headers.
          $entity = $values;
          $columns= max($columns, count($values));
        }
        else {
          // Make sure we don't have more values than columns.
          $values = array_pad(
            array_slice($values, 0, count($columns)), count($columns),
            NULL
          );
          $entity = array_combine($columns, $values);
        }
        $data[$entity[$idf]] = $entity;
        // Process next line.
        $raw_line = array_shift($raw_tsv);
      } while (isset($raw_line));

      // Check for unlabeled columns.
      if (is_int($columns)) {
        $data[''] = range(0, $columns - 1);
      }
    }
    return $data;
  }

  /**
   * {@inheritdoc}
   */
  protected function generateRawData(array $entities_data) :string {
    $raw_data = '';
    // Ignore columns (handled by generateFileHeader()).
    unset($entities_data['']);
    // Write records.
    foreach ($entities_data as $id => $entity) {
      $entity_data = [];
      foreach ($columns as $field) {
        $entity_data[] = str_replace(
          ["\\", "\n", "\t", "\r"],
          ['\\\\', '\n', '\t', '\r'],
          $entity[$field] ?? ''
        );
      }
      $raw_data .= implode("\t", $entity_data) . "\n";
    }
    return $raw_data;
  }

  /**
   * Generates file header.
   *
   * @param string $file_path
   *   Path to the original file.
   *
   * @param array $entities_data
   *   An array of entity data.
   *
   * @return string
   *   The file header.
   */
  protected function generateFileHeader(
    string $file_path,
    array $entities_data
  ) :string {
    $header = '';
    $raw_tsv = file($file_path);
    if (!empty($raw_tsv)) {
      $trash_regex = '/^\\s*$/';
      // Prepare empty lines and comment regex.
      if (!empty($this->configuration['comment_prefix'])) {
        $comment_prefix_len = strlen($this->configuration['comment_prefix']);
        $trash_regex =
          '/^\Q'
          . $this->configuration['comment_prefix']
          . '\E|^\s*$/'
        ;
      }

      // Get and keep existing header.
      $line = '';
      while (isset($line) && preg_match($trash_regex, $line)) {
        $header .= $line;
        $line = array_shift($raw_tsv);
      }

      // Add field names.
      switch ($this->configuration['headers'] ?? '') {
        case 'first':
          // Add columns.
          $header .=
            implode(
              $this->configuration['separator'],
              $entities_data['']
            )
            . "\n"
          ;
          break;

        default:
          break;
      }
    }
    return $header;
  }

}
